//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	BadPixelMap.h --- classes related to bad pixels
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2001-02-19: 1.0.0: - first version from CalAtomsCommon.h
//		UL, 2001-09-14: 1.1.0: - added predicate classes to facilitate
//								 using BadPixelMap in conjunction with
//								 stdlib algorithms
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/BadPixelMap.h,v 1.7 2004/08/04 12:18:14 rsaxton Exp $
//

#ifndef BAD_PIXEL_MAP_H
#define BAD_PIXEL_MAP_H

#include <map>

typedef std::pair<short, short>  PixelLocation;

//
//	---( a bad-pixel )-------------------------------------------------------
//
class BadPixel {
public:
	enum DefectTypes {	INTACT = 0,
						HOT,
						FLICKERING,
						DEAD,
						PIN_HOLE,
						ANOMALOUS
					 };
	BadPixel() {}
	BadPixel(PixelLocation where,
			 unsigned yextent = 1,
			 DefectTypes type = DEAD,
			 bool uplinked = true,
			 double auxAttribute = 0.)
	    	:
			 _where(where),
			 _yextent(yextent),
			 _type(type),
			 _uplinked(uplinked),
			 _auxAttribute(auxAttribute) {}

	PixelLocation	where() const		{ return _where; }
	unsigned		yextent() const		{ return _yextent; }
	DefectTypes		type() const		{ return _type; }
	bool			uplinked() const	{ return _uplinked; }
	double			auxAttribute() const{ return _auxAttribute; }
	virtual unsigned severity() const
				{ errstr<<"A non-OM bad pixel does not have "
							   "a severity value"
							 <<Sas::warning(std::string("BadPixel::severity"));
							  return 0; }

	virtual ~BadPixel() {}

private:
	//	data members
	PixelLocation		_where;		// coordinates
	unsigned			_yextent;	// extent of badness in y
	DefectTypes			_type;		// type of defect
	bool				_uplinked;	// pixel declared "bad" on board
	double				_auxAttribute;
									// additional bad pixel attribute expressed
									// as single real number, e.g. number
									// of saturated impurities for EPN
};


class BadPixelOm : public BadPixel {
public:
	BadPixelOm(BadPixel badPix, unsigned severity = 0)
		  : BadPixel(badPix),		// copy constructor
		   _severity(severity) {}
	~BadPixelOm() {}

	unsigned	severity() const { return _severity; }

private:
	unsigned	_severity;	// OM bad pixels also have a severity
};


//
//	---( a Bad-Pixel map )--------------------------------------------------
//
typedef std::map<PixelLocation, BadPixel, std::less<PixelLocation> >
															BadPixelMapType;
class BadPixelMap : public BadPixelMapType {
public:
	BadPixelMap() {}
		//
		//	PURPOSE:
		//		Instaniate BadPixelMap object
		//

	const BadPixel *isPixelBad(const PixelLocation &where) const;
		//
		//	PURPOSE:
		//		Check if a given pixel is bad.
		//
		//	RETURNS:
		//		Pointer to bad pixel instance if input pixel is among the
		//		the list of bad pixels or 0 otherwise.
		//

	virtual ~BadPixelMap() {}
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//
};

//
//	predicates to find bad pixels with stdlib algorithms
//
template<BadPixel::DefectTypes DefectType>
class isDefect :
	public std::unary_function<const BadPixelMapType::value_type &, bool> {
	public:
		bool operator()(const BadPixelMapType::value_type &bp) const {
			return bp.second.type() == DefectType;
	}
};

template<BadPixel::DefectTypes DefectType>
class isDefectAndUplinked : public isDefect<DefectType> {
public:
	bool operator()(const BadPixelMapType::value_type &bp) const {
		return isDefect<DefectType>::operator()(bp) && bp.second.uplinked();
	}
};

template<BadPixel::DefectTypes DefectType>
class isDefectAndNotUplinked : public isDefect<DefectType> {
public:
	bool operator()(const BadPixelMapType::value_type &bp) const {
		return isDefect<DefectType>::operator()(bp) && !bp.second.uplinked();
	}
};

#endif
