//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalOalUltility.cc	--- various utility function used in the CAL/OAL
//
//	DESCRIPTION:
//		This file contains various utility function used by the CAL/OAL
//
//	CHANGE HISTORY:
//		UL, 1997-11-04: 1.0.0: - skeleton
//		UL, 1998-08-17: 1.1.0: - closed a tiny memory leak in stringColumnRow
//							   - added stringColumnRow with string argument
//		UL, 1999-11-10: 1.2.0: - embedded everything in namespace
//		UL, 2002-06-18: 1.3.0: - added mem_fun_bin_op()
//
//	AUTHOR:
//		U. Lammers, Tue Nov  4 13:57:27 WET 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/CalOalUtility.h,v 1.14 2009/07/06 15:04:45 aibarra Exp $
//

#ifndef CALOAL_UTILITY_H
#define CALOAL_UTILITY_H

#include <functional>
#include <cassert>
#include <cmath>
#include <vector>
#include <cstring>
#include <string>

#include "f90cxx.h"
#include "stringutil.h"
#include "errstr.h"
#include "Dal.h"



namespace CalOalUtility {

template<class T> T SQR(T x) { return x*x; }
template<class T> T CUB(T x) { return x*x*x; }

template<class RealType>
int closestInt(RealType v)
{
	return static_cast<int>(v < 0. ? std::ceil(v - .5) : std::floor(v + .5));
} // closestInt


inline std::string f90string(const char *s, unsigned len)
{
	//	chop-off trailing blanks
	return trim(std::string(s, len));
}


inline char *stringf90(const std::string s, char *f90s, unsigned f90sLen)
{
	assert(f90s);

	std::string str(f90sLen, ' ');
	str.replace(0, s.length()-1, s);
	strncpy(f90s, str.c_str(), f90sLen);

	return f90s;
} // stringf90



typedef std::vector<unsigned> UVec;	// without this dummy declaration
									// compilation of function header below
									// breaks (at least on Solaris/g++ 2.7.2

template<class Type, class DopeT>
inline static void setupDope(DopeT &dope, const Type *vec,
							 std::vector<unsigned> &ext)
{
	dope.addr			= (Pointer)vec;		// address to first element
	for (unsigned i=0; i<ext.size(); ++i) {
		dope.dim[i].extent	= ext[i];		// extent of that dimension
		dope.dim[i].mult	= 1;			// stride; distance between
											// successive elements
		dope.dim[i].lower	= 0;			// lower bound

		for (unsigned j=0; j<i; ++j)
			dope.dim[j].mult *= ext[i];
	}

	dope.offset = 0;
	for (unsigned i=0; i<ext.size(); ++i) {
		dope.dim[i].mult *= sizeof(Type);
		dope.offset += dope.dim[i].mult * dope.dim[i].lower;
	}
	dope.offset = -dope.offset;
} // setupDope()


template<class FromElementType, class ToElementType>
inline void vector2cvec(const std::vector<FromElementType> &v,
						ToElementType *vec, unsigned n)
{
	for (unsigned i=0; i<n; ++i)
		vec[i] = static_cast<ToElementType>(v[i]);
} // vector2cvec

std::string stringColumnRow(const ColumnData *cd, unsigned long row);
	//	this function used to be inlined but it leads to an error
	//	in the CAL - compiler bug?
	//


void stringColumnRowSet(const ColumnData *cd, unsigned long row,
					    const std::string &str);

template<class T>
T optionalAttribute(const Block *block, const std::string &attName,
					T defaultValue);

template<class T>
std::vector<T> &fillVector(const Table *table, const std::string &colName,
						   std::vector<T> &v);

std::vector<std::string> &fillVector(const Table *table,
									 const std::string &colName,
									 std::vector<std::string> &v);

template<class T>
std::vector<T> &fillVector(const Table *table, const std::string &colName,
						   unsigned row, std::vector<T> &v);

template<class T>
std::vector<std::vector<T > > &fillMatrix(const Table *table,
									const std::string &colName,
									std::vector<std::vector<T> > &v);


}

//
//	adapter which allows to combine a stdlib predicate with a pointer
//	to member function and use this a algorithm functor
//
//	Example:
//
//	#include <algorithm>
//	#include <functional>
//	...
//
//	class X {
//	public:
//		X(unsigned i) : _i(i) {}
//		unsigned val() const { return _i; }
//	private:
//		unsigned _i;
//	};
//
//	vector<const X *> v;
//	...
//	vector<const X *>::const_iterator i =
//		find_if(v.begin(), v.end(),
//			bind2nd(mem_fun_bin_op(&X::val, equal_to<unsigned>()), 5));
//
template<typename T, typename R, class BinOp>
class mem_fun_bin_op_t : public std::binary_function<const T*, R, bool> {
public:
	/* AI gcc4 porting
	 * The call to this template in Odf.cc was expecting
	 * a const pointer to a member function. R (T::*p)() const
	 */
	mem_fun_bin_op_t(R (T::*p)() const, const BinOp &op) : _pmf(p), _op(op) {}

	bool operator()(const T *const &p, const R &r) const {
		return _op((p->*_pmf)(), r);
	}

private:
	/* AI gcc4 porting
	 * The call to this template in Odf.cc was expecting
	 * a const pointer to a member function. R (T::*p)() const
	 */
	R (T::*_pmf)() const;
	BinOp	_op;
};
/* AI gcc4 porting
 * The call to this template in Odf.cc was expecting
 * a const pointer to a member function. R (T::*p)() const
 */
template<typename T, typename R, class BinOp>
mem_fun_bin_op_t<T, R, BinOp> mem_fun_bin_op(R (T::*p)() const, const BinOp &op)
{
	return mem_fun_bin_op_t<T, R, BinOp>(p, op);
}

#endif
