//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Image.h --- two-dimensional array class
//
//	DESCRIPTION:
//		Image is a class which represents two dimensional arrays
//		in the absence of the standard library class <valarray>
//
//	CHANGE HISTORY:
//		UL, 1997-12-09: 1.0.0: - skeleton
//		UL, 1999-03-10: 1.1.0: - bug fix in operator[]
//		UL, 2000-01-26: 1.2.0: - added explicit copy constructor
//		UL, 2000-03-15: 1.3.0: - added assignement operator
//		UL, 2000-09-20: 1.4.0: - new operator>> to add Image to Dal Array
//							   - various modifications in constructors
//
//	AUTHOR:
//		U. Lammers, Tue Dec  9 16:27:24 MET 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/Image.h,v 1.10 2009/06/23 13:46:02 aibarra Exp $
//

#ifndef IMAGE_H
#define IMAGE_H

#include <cassert>
/* AI gcc4 porting
 * missing string.h include
 */
#include<string.h>
#include "Dal.h"

typedef float (*Float)(unsigned, unsigned);

template<class ElementType>
class Image {
public:
	//	default constructor
	Image();

	//	construct image of given extents and inital value
	Image(unsigned xExtent, unsigned yExtent, ElementType initialValue = 0);

	//	construct image of given extents and flat vector
	template<class T>
	Image(unsigned xExtent, unsigned yExtent, const T *vector);

	//	construct image of given extents and tor
	typedef const ElementType *FloatFunc (unsigned, unsigned);
	Image(unsigned xExtent, unsigned yExtent, FloatFunc f);

	//	copy constructor/assignment - performs deep copy
	Image(const Image &image);
	Image<ElementType> &operator=(const Image<ElementType> &rhs);

	//	destructor
	~Image() { delete[] _data; }

	unsigned long xExtent() const { return _xExtent; }
	unsigned long yExtent() const { return _yExtent; }
	unsigned long rowSize() const { return _rowSize; }
	unsigned long size() const { return xExtent() * yExtent(); }
	unsigned long memSize() const { return size() * _elemSize; }

	ElementType *operator[](unsigned rowNr) const;
	ElementType *data() const { return _data; }

	Image<ElementType> *window(unsigned x0, unsigned y0,
							   unsigned nx, unsigned ny) const;

	template<class T>
	void fill(const T *vector);

	Image<ElementType> &normalize(ElementType v = 1);

	void resize(unsigned xExtent, unsigned yExtent);
private:
	unsigned long	index(unsigned ix, unsigned iy);
	unsigned long	_xExtent,	//	extent along X
					_yExtent,	//	extent along Y
					_elemSize,	//	size of element;
					_rowSize;	//	_xExtent * _elemSize
	ElementType		*_data;		//	the data as 1-D vector
};


template<class ElementType>
inline void Image<ElementType>::resize(unsigned xExtent, unsigned yExtent)
{
	delete [] _data;
	assert(xExtent > 0 && yExtent > 0);
	_xExtent = xExtent;
	_yExtent = yExtent;

	//	get memory
	_data = new ElementType[xExtent * yExtent];
	assert(_data);

	_elemSize = sizeof(ElementType);
	_rowSize  = _elemSize * _xExtent;
} // Image::resize


template<class ElementType>
inline Image<ElementType>::Image()
	: _xExtent(0), _yExtent(0), _elemSize(0), _rowSize(0), _data(0)
{
} // Image::Image


template<class ElementType>
inline Image<ElementType>::Image(unsigned xExtent, unsigned yExtent,
								 ElementType initialValue) : _data(0)
{
	resize(xExtent, yExtent);
	for (unsigned long n=0; n<xExtent * yExtent; _data[n++] = initialValue) ;
} // Image::Image


template<class ElementType>
template<class T>
inline Image<ElementType>::Image(unsigned xExtent, unsigned yExtent,
								 const T *vector) : _data(0)
{
	resize(xExtent, yExtent);
	fill(vector);
} // Image::Image


template<class ElementType>
template<class T>
inline void Image<ElementType>::fill(const T *v)
{
	unsigned N = xExtent() * yExtent();
	for (unsigned n=0; n<N; ++n)
		_data[n] = static_cast<ElementType>(v[n]);
} // Image::fill()


template<class ElementType>
inline Image<ElementType>::Image(const Image &image)
{
	_data = 0;
	this->operator=(image);
} // Image::Image

template<class ElementType>
Image<ElementType> &Image<ElementType>::operator=(
											const Image<ElementType> &rhs)
{
	resize(rhs.xExtent(), rhs.yExtent());
	memcpy(_data, rhs.data(), _xExtent * _yExtent * _elemSize);

	return *this;
} // Image


template<class ElementType>
inline unsigned long Image<ElementType>::index(unsigned ix, unsigned iy)
{
	return (_yExtent * ix) + iy;
} // Image::index


template<class ElementType>
inline Image<ElementType>::Image(unsigned xExtent, unsigned yExtent,
								 FloatFunc f) : _data(0)
{
	resize(xExtent, yExtent);
	for (unsigned ix=0; ix<xExtent; ++ix)
		for (unsigned iy=0; iy<yExtent; ++iy)
			_data[index(ix, iy)] = static_cast<ElementType>((*f)(ix, iy));
} // Image


template<class ElementType>
inline ElementType *Image<ElementType>::operator[](unsigned rowNr) const
{
	assert(rowNr < _xExtent);
	return _data + _yExtent * rowNr;
} // Image::operator[]


template<class ElementType>
inline Image<ElementType> *Image<ElementType>::window(unsigned x0,
													  unsigned y0,
													  unsigned nx,
													  unsigned ny) const
{
	Image<ElementType> *subImage = new Image<ElementType>(nx, ny);

	//	inefficient - to be improved
	for (unsigned ix=0; ix<nx; ++ix)
		for (unsigned iy=0; iy<ny; ++iy)
			(*subImage)[ix][iy] = (*this)[x0+ix][y0+iy];

	return subImage;
} // Image::window

template<class ElementType>
inline Image<ElementType> &Image<ElementType>::normalize(ElementType v)
{
	unsigned N = _xExtent * _yExtent;
	ElementType sum = 0;
	for (unsigned n=0; n<N; sum+=_data[n++]) ;
	ElementType c = v/sum;
	for (unsigned n=0; n<N; ++n)
		_data[n] *= c;

	return *this;
} // Image::normalize()

typedef Image<float>	SImage;
typedef Image<double>	DImage;
typedef Image<unsigned>	UImage;
typedef Image<int>		I32Image;
typedef Image<short>	I16Image;

template<class Type>
DataSet *operator<<(DataSet *set, const Image<Type> &image)
{
    std::vector<unsigned long> extents(2);
	extents[0] = image.xExtent();
	extents[1] = image.yExtent();

	Array *a = set->addArray("PSF", Dal::Real32, extents);
    unsigned nElements = a->elements();
    ArrayData *ad = a->data();
    real32 *d = ad->real32Data();
    for (unsigned n=0; n<nElements; ++n)
		d[n] = image.data()[n];
	delete ad;

	return set;
} // operator<<

template<class Type>
Array *operator>>(Array *array, Image<Type> &image)
{
	Array &a = *array;
	image.resize(a.size()[0], a.size()[1]);
	ArrayData *ad = a.data();
	switch(a.dataType()) {
		case Dal::Int8:		image.fill(ad->int8Data()); break;
		case Dal::Int16:	image.fill(ad->int16Data()); break;
		case Dal::Int32:	image.fill(ad->int32Data()); break;
		case Dal::Uint16:	image.fill(ad->uint16Data()); break;
		case Dal::Uint32:	image.fill(ad->uint32Data()); break;
		case Dal::Real32:	image.fill(ad->real32Data()); break;
		case Dal::Real64:	image.fill(ad->real32Data()); break;
		default: break;
	}

	delete ad;
	return array;
} // operator>>

#endif
