//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	PositionLocator.h	--- classes related to objection position inquiry
//
//	DESCRIPTION:
//		An abstract class PositionLocator + auxiliary classes are declared
//		which provide methods to inquire the position of an object as a
//		function of time.
//
//	CHANGE HISTORY:
//		UL, 2000-11-02: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Tue Aug 11 15:55:34 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/PositionLocator.h,v 1.5 2004/08/04 12:18:14 rsaxton Exp $
//

#ifndef POSITION_LOCATOR_H
#define POSITION_LOCATOR_H

#include <iostream>

#include "Celestial.h"
#include "STime.h"
#include "errstr.h"

struct StateVector {
	StateVector(double x = 0., double y = 0., double z = 0.,
				double vx = 0., double vy = 0., double vz = 0.,
				double rev = 0.)
		: position(x, y, z), velocity(vx, vy, vz),
		  revolution(rev) {}

	StateVector(const double v[], double rev = 0.)
		: position(v[0], v[1], v[2]), velocity(v[3], v[4], v[5]),
		  revolution(rev) {}

	double operator[](unsigned n) const
	{
		if (n>=6)
			errstr<<"Index out of bounds: "<<n<<" - must be 0<=n<6"
				<<Sas::error("IndexOutOfBounds");

		if (n==0) return position.x();
		if (n==1) return position.y();
		if (n==2) return position.z();
		if (n==3) return velocity.x();
		if (n==4) return velocity.y();

		return velocity.z();
	}


	CartesianVector	position, velocity;
		// object position and velocity in Cartesian, geocentric
		// coordinates - units: km and km/s - frame is aligned
		// with mean equatorial reference frame of equinox J2000
	double revolution;
};

inline std::ostream &operator<<(std::ostream &os, const StateVector &stateVec)
{
	os.precision(8);
	os<<"StateVector(x="<<stateVec.position.x()<<", y="
	  <<stateVec.position.y()<<", z="<<stateVec.position.z()
	  <<", vx="<<stateVec.velocity.x()<<", vy="<<stateVec.velocity.y()
	  <<", vz="<<stateVec.velocity.z()<<" revolution="<<stateVec.revolution;
	return os;
}

class PositionLocator {
public:
	PositionLocator(const STime &baseTime) : _baseTime(baseTime) {}
		//
		//	PURPOSE:
		//		Construct PositionLocator object with a given basetime
		//
		//	ARGUMENTS:
		//		baseTime	: basetime to which offsets given to
		//					  PositionLocator::stateVector(time) will be
		//					  added to obtain absolute time
		//

	const STime &baseTime() const { return _baseTime; }
	void baseTime(const STime &time) { _baseTime = time; }
		//
		//	PURPOSE:
		//		Retrieve/redefine current base time
		//

	virtual StateVector stateVector(double time) const = 0;
	virtual StateVector stateVector(const STime &time) const = 0;

	virtual ~PositionLocator() {}

private:
	//	data
	STime	_baseTime;
};

#endif
