//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	StringPool.cc	--- implementation of class StringPool
//
//	DESCRIPTION:
//		see StringPool.h
//
//	CHANGE HISTORY:
//		UL, 1998-05-28: 1.0.0: - first version
//		UL, 1998-10-02: 1.1.0: - return -> ALTRETURNV line #264
//		UL, 1998-08-05: 1.2.0: - input/output streams are no longer
//								 data members of StringPool class; use
//								 cin and errstr (from error package) instead
//		UL, 2000-09-29: 1.3.0: - more robustness: do not bail out on
//								 illegal characters
//
//	AUTHOR:
//		U. Lammers, Thu May 28 17:44:55 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/caloalutils/src/StringPool.cc,v 1.12 2009/06/23 13:46:02 aibarra Exp $
//

#include <cctype>

#include "StringPool.h"
#include "errstr.h"

const unsigned VEC_SIZE = ('z'-'a'+1) + 10 + 1 + 1 + 1  + 1;
//                           [a-z]     [0-9] _   -  " "  anythingelse

static unsigned index(char c)
{
	if ((c >= 'a' && c <= 'z') || (c >= 'A' && c<= 'Z'))
		return tolower(c) - 'a';
	else if (isdigit(c))
		return c - '0' + 'z'-'a' + 1;
	else if (c == '_')
		return VEC_SIZE - 4;
	else if (c == '-')
		return VEC_SIZE - 3;
	else if (c == ' ')
		return VEC_SIZE - 2;
	else
		errstr<<"Illegal character `"<<c<<"' in string"
			  <<Sas::warning("IllegalCharacter");

	return VEC_SIZE-1;
} // index()


class NodeData {
public:
	friend class StringPool;

	NodeData(char c = '*', bool term = false, unsigned count = 0)
			: _c(c), _term(term), _count(count)
	{
		_single = 0;
		_next = 0;
	}

	char c() { return _c; }

	void trueTerm() { _term = true; }
		//	make node as a terminal node

	bool term() const { return _term; }
		//	Is the node a terminal node?

	void incrCount() { ++_count; }
		//	increment reference cout

	unsigned count() const { return _count; }
		//	get value of reference counter

	NodeP registerChild(char c)
	{
		unsigned indx = index(c);
		if (_single && tolower(_single->c()) != tolower(c)) {
			_next = new std::vector<NodeP>(VEC_SIZE);
			for (unsigned n=0; n<VEC_SIZE; (*_next)[n++] = 0) ;
			(*_next)[index(_single->c())] = _single;
			_single = 0;
			return registerChild(c);
		}

		NodeP &nd = _next ? (*_next)[indx] : _single;
		if (nd)
			nd->incrCount();
		else
			nd = new NodeData(c, false, 1);

		return nd;
	}
	NodeP child() const { return _single; }
	NodeP child(char c) const
	{
		if (_next)
			return (*_next)[index(c)];

		return (_single && tolower(_single->c()) == tolower(c)) ? _single : 0;
	}

	virtual ~NodeData()
	{
		if (_next)
			for (unsigned n=0; n<_next->size(); delete (*_next)[n++]) ;
		delete _next; _next = 0;
		delete _single; _single = 0;
	}

private:
	char				_c;		//	what is the character
	bool				_term;	//	is this a terminal node?
	unsigned			_count;	//  element of how many strings in pool
	NodeP				_single;//	point to only child
	std::vector<NodeP>	*_next;	//	vector with pointers to children
}; // NodeData


void StringPool::init()
{
	_root = new NodeData;
} // StringPool::init()



StringPool::StringPool(Verbosity verbosity)	: _verbosity(verbosity)
{
	init();
} // StringPool()


void StringPool::addString(const char *s)
{
	this->push_back(s);
	if (!s || !*s)
		return;

	NodeP n = _root;
	for (const char *p = s; *p; ++p)
		n = n->registerChild(*p);

	n->trueTerm();		//	mark node as terminal
	_root->incrCount();	//	a new entry

} // StringPool::addString()


StringPool::StringPool(const char *stringVec[], Verbosity verbosity)
				: _verbosity(verbosity)
{
	init();
	if (!stringVec || !*stringVec)
		return;

	const char **p = stringVec;
	while (*p)
		addString(*p++);
} // StringPool::StringPool()


StringPool::StringPool(const std::vector<std::string> &stringVec,
					   Verbosity verbosity)
			: _verbosity(verbosity)
{
	init();
	for (unsigned n=0; n<stringVec.size(); ++n)
		addString(stringVec[n].c_str());
} // StringPool::StringPool()


void StringPool::completeSingle(const NodeP n, std::string &matchString) const
{
	if (n) {
		matchString += n->c();
		if (!n->term()) {
			NodeP ch = n->child();
			if (!ch)
				errstr<<"More than 1 completion possible\n"
					  <<Sas::error("AmbigiousCompletion");
			completeSingle(ch, matchString);
		}
	}
} // StringPool::completeSingle()



std::vector<std::string> StringPool::completeMulti(const NodeP node) const
{
	std::vector<std::string> v(node->count());
	if (node->count() == 1)
		completeSingle(node, v[0]);
	else
		if (node->_single) { // there is only one child
			std::vector<std::string> v2 = completeMulti(node->child());
			for (unsigned n=0; n<v.size(); ++n)
				v[n] = node->c() + v2[n];
		} else {
			unsigned nStrings = 0;
			for (unsigned n=0; n<node->_next->size(); ++n)
				if ((*node->_next)[n]) {
					std::vector<std::string> v2 =
											completeMulti((*node->_next)[n]);
					for (unsigned nn=0; nn<v2.size(); ++nn)
						v[nStrings + nn] = node->c() + v2[nn];
					nStrings += v2.size();
				}
		}

	return v;
} // StringPool::completeMulti()


bool StringPool::match(const NodeP n, const char *p, std::string &matchString,
					   NodeP &lastMatched) const
{
	lastMatched = 0;
	if (*p)
		if (NodeP c = n->child(*p)) {
			matchString += c->c();
			return match(c, ++p, matchString, lastMatched);	//	keep trying
		} else
			return false;		//	no match!
	else
		if (n->term())
			return true;		//	exact match!
		else
			if (n->count() == 1) {
				completeSingle(n->child(), matchString);
				return true;	//	unique completion
			} else {
				lastMatched = n;
				return false;	//	ambigious
			}

	return false;
} // StringPool::match()


std::string StringPool::expand(const std::string &keyString) const
{
	std::string matchString = "";
	NodeP lastNode;
	/* AI gcc4 porting
	 * suggest explicit braces to avoid ambiguous 'else'
	 */
	if (!match(_root, keyString.c_str(), matchString, lastNode))
	{
		if (lastNode) {
			//	ambigious completion
			if (_verbosity != StringPool::SILENT) {
				errstr<<"Ambigious completion:\n";
				std::vector<std::string> v = completeMulti(lastNode);
				for (unsigned n=0; n<v.size(); ++n)
					errstr<<matchString<<"|"<<v[n].substr(1, v[n].length()-1)
					<<std::endl;
				if (_verbosity == StringPool::ABORT)
					Sas::error("Abort");
				else {
					std::string s;
					errstr<<"Please give more characters: ";
					std::cin>>s;
					return expand(matchString + s);
				}
			}
		} else
			if (_verbosity != StringPool::SILENT) {
				errstr<<"No match - best match was: `"<<matchString<<"'\n";
				switch(_verbosity) {
				case WARN:
					Sas::warning("NoMatch"); break;
				case ABORT:
					Sas::error("NoMatch"); break;
				case CORRECT: {
					std::string str;
					errstr<<"No match - input again: ";
					std::cin>>str;
					return expand(str);
					break;
				}
				default: break;
				}
			}
	}

	return matchString;
} // StringPool::expand()


const std::string &StringPool::operator[](const std::string &keyString) const
{
	static std::string s = expand(keyString);
	for (unsigned n=0; n<this->size(); ++n)
		if (s == (*this)[n])
			return (*this)[n];

	return s;
} // StringPool::operator[]


StringPool::~StringPool()
{
	delete _root; _root = 0;
} // StringPool::~StringPool()
